From elpi.apps.tc.tests.importOrder Require Import f3f.
From elpi.apps.tc.tests.importOrder Require Import f2a.
From elpi.apps.tc.tests.importOrder Require Import f2b.
From elpi.apps.tc.tests.importOrder Require Import f3c.
From elpi.apps.tc.tests.importOrder Require Import f3d.
From elpi.apps.tc.tests.importOrder Require Import f3g.

Elpi SameOrderImport.